#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.12,<3.13"
# dependencies = [
# "licofage",
# ]
# ///
import stat
from argparse import ArgumentParser, FileType
from licofage.argmisc import Formatter
from licofage.kit import *
from os import environ
import datetime
from pathlib import Path as P
from tempfile import TemporaryDirectory
from shutil import make_archive, copy

today = datetime.date.today().strftime("%Y-%m-%d")
parser = ArgumentParser(
    prog="gen_factcomplmultway",
    description="""Générateur de Complexité en Facteur

Génère un document .qmd pour calculer la complexité en facteur d'une substitution donnée.

La substitution est supposée Pisot. Le calcul utilise l'outil Walnut.

Utilisation: $ ./gen_factcomplmultfact tri '01/02/0'""",
    formatter_class=Formatter,
)

parser.add_argument(
    "-o",
    "--output",
    type=FileType("w"),
    default=None,
    help="output file (default to [name].qmd)",
)
# parser.add_argument(
#     "-z",
#     "--zip",
#     action="store_true",
#     help="create zipped Walnut instance instead of notebook",
# )
# parser.add_argument(
#     "-u", "--unroll", action="store_true", help="unroll Walnut scripts"
# )
parser.add_argument(
    "-v", "--verbose", action="store_true", help="display more messages"
)
parser.add_argument(
    "-S", "--stats", action="store_true", help="display statistics about computations"
)
parser.add_argument("name", help="numeration system and fixpoint naming")
parser.add_argument("subst", help="substitution considered")
parser.add_argument("fullname", help="full name of the studied sequence (for the title)", default=None
)
args = parser.parse_args()

if args.zip:
    tmpdir = TemporaryDirectory()
    basedir = P(tmpdir.name) / args.name
    basedir.mkdir()
    walnutdir = basedir / "Walnut"
    walnutdir.mkdir()
    for s in "Result,Command Files,Custom Bases,Automata Library,Word Automata Library".split(
        ","
    ):
        (walnutdir / s).mkdir()
    walnut_out = open(walnutdir / "Command Files" / "gen_factcomp1.txt", "w")
    py_out = open(basedir / "prelim.py", "w")
    goscr = basedir / "go"
    goscr_out = open(goscr,"w")

if not args.zip and args.output is None:
    args.output = open(f"{args.name}.qmd", "w")

if args.fullname is None:
    args.fullname = args.name.title()

out = args.output


def to_md(s):
    # Injecte dans le .qmd du texte en Markdown
    if not args.zip:
        out.write(s)


def to_py(s):
    # Injecte dans le .qmd du code python 
    # ou l'ajoute dans le script python du zip
    if args.zip:
        py_out.write(s)
        py_out.write("\n")
    else:
        out.write(f"""
```{walnut}
%%python
{s}
```
""")

genv = dict()

def do_py(s, ss=None):
    # Execute et écrit dans le .qmd ou le script du zip le code python {s} et {ss}
    global genv
    if args.unroll:
        exec(s, genv)
    to_py(f"{s}\n{ss}" if ss is not None else s)

def to_wal(s):
    # Injecte le texte sous forme de cellule walnut dans le .qmd
    # ou l'ajoute au code walnut du zip
    if args.zip:
        walnut_out.write(s)
        walnut_out.write("\n")
    else:
        out.write(f"""
```{walnut}
{s}
```
""")


def to_pyscript(sname, py):
    # Injecte dans le .qmd du code python qui génère un script de walnut et ajoute une cellule de walnut pour appeler le code qui sera généré
    # éventuellement execute le code générateur et insère directement le code dans une cellule Walnut
    global genv
    if args.unroll:
        genv["script"] = []
        exec(py, genv)
        to_wal("\\n".join(genv["script"]))
    else:
        to_py(f"""script = []
{py}
writefile(f"{sname}", "\\n".join(script))""")
        to_wal(f"""load {sname}.txt;""")



if args.zip:
    goscr_out.write("""#!/bin/sh
export WALNUT_MEM=64g
export WALNUT_HOME=$(pwd)/Walnut
uv run --python 3.12 --with licofage prelim.py
cd $WALNUT_HOME
echo "load gen_factcomp1.txt;" | java -Xmx$WALNUT_MEM -jar $WALNUT_JAR 
""")


to_md(f"""---
title: Complexité en Facteur de {args.fullname} 
date: '{today}'
format:
  html:
    code-fold: false
    embed-resources: true
    toc: true
    toc-location: left
    number-sections: true
    syntax-definitions: 
        - walnut.xml
  ipynb: default 
execute:
  cache: true
  keep-ipynb: true
  allow_errors: true
jupyter: walnut
---
""")

walnut = "{walnut}"
name = args.name
base = f"msd_{name}"
word = name.title()

to_md(f"""
## Construction du système de numération

On défini la substitution, le système de numération et le point fixe associé.
""")
do_py(f"""from licofage.kit import *
import os
setparams({repr(args.verbose)}, {repr(args.stats)}, os.environ["WALNUT_HOME"])

s = subst({repr(args.subst)})
ns = address(s, "{name}")""", 
"""ns.gen_ns()
ns.gen_word_automaton()""")

to_md(f"""
## Prédicats de base

`feq_{name}(i,j,n)` : accepte (i,j,n) si les facteurs (i,n) et (j,n)
sont égaux  
`first_{name}(i,n)` : accepte (i,n) si c'est un facteur nouveau  
`last_{name}(i,n)` : accepte (i,n) si c'est un facteur nouveau et qu'aucun facteur nouveau ne se situe après lui  
""")

to_wal(f"""def cut_{name} "?{base} i<=u & j<=v & u+j=v+i & u<n+i & v<n+j":
def feq_{name} "?{base} ~(Eu,v $cut_{name}(i,j,n,u,v) & {word}[u]!={word}[v])":
def first_{name} "?{base} Aj $feq_{name}(i,j,n) => i<=j":
def last_{name} "?{base} $first_{name}(i,n) & Ak k>i => ~$first_{name}(k,n)": """)

to_md(f"""
## Construction de la complexité en facteur par itération 

`next_block_{name}(i,j,n)` : (i,n) est un facteur non nouveau et (j,n) est le premier facteur nouveau qui suit. Si un tel facteur (j,n) n'existe pas, j=i  
`block_max_right_{name}(i,j,n,l)` : il y a un bloc de l facteurs nouveau entre (i,n) et (j,n) et il est maximal à droite  
Suite d'automate / prédicat (j entier non nul) `compte_j_{name}(e,n,p)` : entre (0,n) et (e,n), il y a p facteurs nouveau répartis dans j blocs ou moins  
`compl_fact_{name}_V1(n,p)` : complexité en facteur synchronisée
""")

to_pyscript("factcompl_via_iter",f"""
script.append(f'def last_{name} "?{base} $first_{name}(i,n) & Ak k>i => ~$first_{name}(k,n)";')
script.append(f'''def block_max_right_{name} "?{base} (~$first_{name}(i,n) & k=0 & j=i) 
        | ($first_{name}(i,n) 
            & (Ap (i<=p & p<=j) => $first_{name}(p,n))    
            & ~$first_{name}(j+1,n) 
            & (i+k = j+1))";''') 
script.append(f'''def next_block_{name} "?{base} (i<j & ~$first_{name}(i,n) & $first_{name}(j,n) & 
    (Ap (i<p & p<j) => ~$first_{name}(p,n)))
    | (Ek $last_{name}(k,n) & k<i & i=j)";''')                                
script.append(f'def compte_0_{name} "?{base} $block_max_right_{name}(0,e,p,n)":')
script.append(f'def compte_1_{name} "?{base} Es,q $compte_0_{name}(s,n,q)&(($next_block_{name}(s+1,s+1,n) & p=q & s+1=e+1)|(Ek s+1<k & $next_block_{name}(s+1,k,n) & El $block_max_right_{name}(k,e,l,n) & p=q+l))";')
script.append('eval test_egal_{name} "?{base} Ae,n,p $compte_1_{name}(e,n,p) <=> $compte_0_{name}(e,n,p)";')
j=1
""")

to_pyscript("factcompl_via_iter", f"""
p_test = P(os.environ["WALNUT_HOME"])/"Result"/"test_egal_{name}.txt"
flag = 0
with p_test.open("r") as f_test:
    res = f_test.read()
    if "false" in res :
        flag = 1
    elif "true" in res:
        flag = 2
    else :
        print("Un problème est survenu")
if flag == 1: 
    j+=1
    script.append(f'def compte_{{j}}_{name} "?{base} Es,q $compte_{{j-1}}_{name}(s,n,q)&(($next_block_{name}(s+1,s+1,n) & p=q & s+1=e+1)|(Ek s+1<k & $next_block_{name}(s+1,k,n) & El $block_max_right_{name}(k,e,l,n) & p=q+l))";')
    script.append(f'eval test_egal_{name} "?{base} Ae,n,p $compte_{{j-1}}_{name}(e,n,p) <=> $compte_{{j}}_{name}(e,n,p)";')
    print("Exécutez les deux dernières cellules une fois de plus.")
elif flag == 2 :
    script.append(f'def compl_fact_{name}_V1 "?{base} Ee $compte_{{j}}_{name}(e,n,p)";')
    print("OK")""")


to_md(f"""
## Vérification inductive de l'automate

`new_first_{name}(i,n)` : accepte le facteur (i,n) si il est nouvellement nouveau  
`NEW_{name}[n]` : nombre de facteur nouvellement nouveau de longueur n+1  
""")

to_wal(f"""
def new_first_{name} n "?{base}  ~$first_{name}(i,n) & $first_{name}(i,n+1)":
""")

to_md(f"""
On applique le semigroup-trick.

```{walnut}
%SGT new_first_{name} {name} NEW_{name}
```
""")

to_wal(f"""
eval check_zero "?{base} $compl_fact_{name}_V1(0,1)":
eval check_rec "?{base} An,p,q ($compl_fact_{name}_V1(n,p) & $compl_fact_{name}_V1(n+1,q)) => p+NEW_{name}[n] = q":
""")


to_md(f"""
## Construction de la complexité en facteur à l'aide de NEW
### Récupérer le morphisme associé à NEW
""")

do_py("""
from pathlib import Path as P
import os
from licofage.kit import subst, address

def readDFAO(f):
    # lecture de tous les états, de leurs transitions et sortie
    l = f.readline()
    dfa = dict()
    outputs = dict()
    while l != '':
        etat_data = l.split()
        etat = int(etat_data[0].strip()) #on récupère le nom de l'etat
        outputs[etat] = int(etat_data[1].strip())
        transition = dict()
        l = f.readline()
        while "->" in l:
            tr = l.split("->")
            transition[int(tr[0].strip())] = int(tr[1].strip())
            l = f.readline()
        dfa[etat] = transition
        l = f.readline()

    return (dfa,outputs)

def DFAtoSubst(dfa):
    subst = ""
    for etat in sorted(dfa.keys()):
        subst += chr(etat+97)+" -> "
        for lect in sorted(dfa[etat].keys()):
            subst += chr(dfa[etat][lect]+97)
        subst += " , "
    return subst[:-2] ## pour la virgule ajoutée en trop

def DFAOtoMorph(dfa,outputs):
    subst = DFAtoSubst(dfa)
    w = dict()
    for key in outputs.keys():
        w[chr(key+97)] = outputs[key]
    return subst, w

def getSubst(p:P):
    with p.open('r') as f:
        # atteindre le premier état
        l = f.readline()
        while l != '\\n':
            l = f.readline()
        (dfa, outputs) = readDFAO(f)
    return DFAOtoMorph(dfa, outputs)
""")


to_md(f"""
### Calculer la complexité en facteur à l'aide de NEW

`sumNEW(n,s)` : s est la somme des n-1 premières valeurs de NEW
`compl_fact_{name}_V2(n,p) : complexité en facteur synchronisée (obtenue à l'aide de NEW)
""")

do_py(f"""
new_subst, w = getSubst(P(os.environ["WALNUT_HOME"])/"Word Automata Library"/"NEW_{name}.txt")
new = subst(new_subst)
new_w = address(new, '{name}',**w)
(new_w-ns).gen_dfa("sumNEW_{name}")
""")

to_wal(f"""
def compl_fact_{name}_V2 "?{base} Eq $sumNEW_{name}(n,q) & p=q+1":
""")

to_md(f"""
### Vérification du résultat

`compl_fact_{name}_V2` est calculée à l'aide de `NEW`. On ne peut pas utiliser la technique précédente pour vérifier notre prédicat. Par contre, on peut confronter `compl_fact_{name}_V1` et compl_fact_{name}_V2 car elle sont issues de deux méthodes différentes.
""")

to_wal(f"""
eval check_V2 "?{base} An,p $compl_fact_{name}_V1(n,p) <=> $compl_fact_{name}_V2(n,p)":
""")


if args.zip:
    goscr_out.write(f"""cat << EOF | tr '\\n' '\\0' | tar cvzf ../factcompl.tar.gz --null -T -
Custom Bases/{base}.txt
Custom Bases/{base}_addition.txt
Word Automata Library/{word}.txt
Automata Library/feq_{name}.txt
Automata Library/first_{name}.txt
Result/new_first_{name}.mpl
Word Automata Library/compl_{name}.txt
EOF\n""")
    goscr.chmod(goscr.stat().st_mode | stat.S_IEXEC)
    walnut_out.close()
    py_out.close()
    goscr_out.close()
    make_archive(f"{args.name}", "zip", tmpdir.name, args.name)
    tmpdir.cleanup()